--  This software is distributed under the following license know as
--  The modified BSD license
--
--  Redistribution and use in source and binary forms, with or without
--  modification, are permitted provided that the following conditions are
--  met:

--      (1) Redistributions of source code must retain the above copyright
--      notice, this list of conditions and the following disclaimer. 

--      (2) Redistributions in binary form must reproduce the above copyright
--      notice, this list of conditions and the following disclaimer in
--      the documentation and/or other materials provided with the
--      distribution.  
    
--      (3)The name of the author may not be used to
--      endorse or promote products derived from this software without
--      specific prior written permission.

--  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
--  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
--  DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
--  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
--  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
--  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
--  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
--  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
--  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
--  POSSIBILITY OF SUCH DAMAGE.

with Scalada.Intervals; use Scalada.Intervals;
with Scalada.Scales_Repository; use Scalada.Scales_Repository;
with Scalada.Notes; use Scalada.Notes;
with Ada.Text_Io; use Ada.Text_Io;
with Ada.Exceptions; use Ada.Exceptions;

package body Scalada.Scale_Finder is

   type Permutations is array(Positive range <>) of Matching_Scale;

   -----------------------
   -- Utility functions --
   -----------------------
   function Extract_Qualities(Notes : in Set_Of_Notes) return Quality_Interval_array is
      Qualities : Quality_Interval_Array(Notes'Range);
   begin
      for index in Notes'Range loop
         Qualities(Index) := Quality_Of(Notes(Index), Notes(Notes'First));
      end loop;

      return Qualities;
   end Extract_Qualities;

   function Extract_Notes(Chords : Chords_Array) return Set_Of_Notes is
      type Present_Notes_array is array(Notes.Note) of Boolean;
      Present_Notes : Present_Notes_Array := (others => False);
      Number_Of_Notes : Natural := 0;
   begin
      -- for each chord provided
      for Index in Chords'Range loop
     Declared_Notes:
         declare
            -- get notes
            Contained_Notes : Set_Of_Notes := Get_Notes(Chords(Index).Chord_Structure, Chords(Index).Tonic);
         begin
            -- mark presence
            for Note_Index in Contained_Notes'Range loop
               if Present_Notes(Contained_Notes(Note_Index)) = False then
                  Number_Of_Notes := Number_Of_Notes + 1;
                  Present_Notes(Contained_Notes(Note_Index)) := True;
               end if;

            end loop;
         end Declared_Notes;
      end loop;

      -- build a set of notes from present notes
      declare
         Set_To_Return : Set_Of_Notes(1..Number_Of_Notes);
         Cursor : Natural := 1;
      begin
         for Index in Present_Notes'Range loop
            if Present_Notes(Index) = True then
               Set_To_Return(Cursor) := Index;
               Cursor := Cursor + 1;
            end if;
         end loop;

         return Set_To_Return;
      end;
   end Extract_Notes;

   function Get_Permutations(Notes : in Set_Of_Notes) return Permutations is
      Returned_Permutations : Permutations(Notes'Range);
      type Mod_Index is new Positive range Notes'Range;
   begin
      for Index in Returned_Permutations'Range loop
         declare
            Permuted_Set : Set_Of_Notes(Notes'Range);
            --Prepare to loop from the index-th element
            Set_Index : Mod_Index := Mod_Index(Index);
         begin
            for Permuted_Set_Index in Permuted_Set'Range loop
               Permuted_Set(Permuted_Set_Index) := Notes(Positive(Set_Index));
               if (Set_Index = Mod_Index'Last) then
                 Set_Index := Mod_Index'First;
               else
                  Set_Index := Set_Index + 1;
               end if;
            end loop;

            -- Build quality intervals from set and assign
            Returned_Permutations(Index) := Matching_Scale'(Permuted_Set(Permuted_Set'First), Create("", Extract_Qualities(Permuted_Set)));
         end;
      end loop;

      return Returned_Permutations;
   end Get_Permutations;
   
   ---------------------
   -- Matching_Scales --
   ---------------------
   function Matching_Scales (Included_Notes : in Set_Of_Notes) return Matching_Scales_Array is
      Scales : Scales_Array := Scales_Repository.Get_Registered_Scales;
      -- Maximal solution array
      Array_To_Return : Matching_Scales_Array(1..Scales'Length);

      -- First find all permutations of the set of notes to have all quality intervals
      Tested_Permutations : Permutations := Get_Permutations(Included_Notes);

      Return_Index : Positive := 1;

   begin
      for Index in Tested_Permutations'Range loop
         for Scales_Index in Scales'Range loop
            if Equals(Tested_Permutations(Index).Matching_scale, Scales(Scales_Index)) = True then
               Array_To_Return(Return_Index) := Matching_Scale'(Tested_Permutations(Index).Tonic, Scales(Scales_Index));
               Return_Index := Return_Index + 1;
            end if;
         end loop;
      end loop;

      return Array_To_Return(1..Return_Index-1);

   end Matching_Scales;

   -----------------------
   -- Containing_Scales --
   -----------------------
   function Containing_Scales(Included_Notes : in Set_Of_Notes) return Matching_Scales_Array is
      Scales : Scales_Array := Scales_Repository.Get_Registered_Scales;

      -- Maximal solution array
      Array_To_Return : Matching_Scales_Array(1..Scales_Repository.Number_Of_Scales * (Note'Pos(Note'Last) - Note'Pos(Note'First)));

      -- First find all permutations of the set of notes to have all quality intervals
      Tested_Permutations : Permutations := Get_Permutations(Included_Notes);

      Return_Index : Positive := 1;
   begin
      for Index in Tested_Permutations'Range loop
         for Scales_Index in Scales'Range loop
            if Contains(Scales(Scales_Index), Tested_Permutations(Index).Matching_Scale) = True then
               Array_To_Return(Return_Index) := Matching_Scale'(Tested_Permutations(Index).Tonic, Scales(Scales_Index));
               Return_Index := Return_Index + 1;
            end if;
         end loop;
      end loop;

      return Array_To_Return(1..Return_Index-1);
   end Containing_Scales;

   -----------------------
   -- Containing_Scales --
   -----------------------
   function Containing_Scales(Included_Chords : in Chords_Array) return Matching_Scales_Array is
   begin
      return Containing_Scales(Extract_Notes(Included_Chords));
   end Containing_Scales;

end Scalada.Scale_Finder;
